#if !defined(LKT_OS_H__)
#define LKT_OS_H__

#if defined(__cplusplus)
/* If this file is included by a C++ file, noreturn is not defined... */
#define noreturn __attribute__((noreturn))
extern "C" {
#endif

#include <lektor/lktconfig.h>

/* This header contains all plateform and OS specific functions, this is the
 * ugly portability and compatibility layer... */

/* Include headers, they are present because the configure checked for them. */
#include <stdnoreturn.h>
#include <inttypes.h>
#include <locale.h>
#include <memory.h>
#include <regex.h>
#include <stddef.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <limits.h>
#include <stdbool.h>
#include <stdarg.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <time.h>
#include <ctype.h>
#include <fcntl.h>

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/time.h>

/* Use the endian.h and not the one computed at configure time */
#if defined(HAVE_ENDIAN_H) && HAVE_ENDIAN_H == 1
#include <endian.h>

#if !(__BYTE_ORDER == __LITTLE_ENDIAN || __BYTE_ORDER == __BIG_ENDIAN)
#error "Unsuported endiannes"
#endif

#endif /* End of endiannes stuff */

#if defined(__linux__)
#include <linux/limits.h>
#endif

typedef void (*function_ptr)(void);

/* Attributes */
#define UNUSED               __attribute__((unused))
#define UNUSED_FUNCTION      UNUSED
#define CONSTRUCTOR_FUNCTION __attribute__((constructor)) static void
#define DESTRUCTOR_FUNCTION  __attribute__((destructor)) static void
#define EXIT_FUNCTION        noreturn void
#define PRIVATE_FUNCTION     UNUSED_FUNCTION static inline

/* Compilers */
#if defined(__clang__)
#define LKT_PUSH                     _Pragma("clang diagnostic push")
#define LKT_IGNORE_WARN_FMT_SECURITY _Pragma("clang diagnostic ignored \"-Wformat-security\"")
#define LKT_IGNORE_WARN_WRITE_STRING _Pragma("clang diagnostic ignored \"-Wwrite-strings\"")
#define LKT_POP                      _Pragma("clang diagnostic pop")

#elif defined(__GNUC__) || defined(__GNUG__)
#define LKT_PUSH                     _Pragma("GCC diagnostic push")
#define LKT_IGNORE_WARN_FMT_SECURITY _Pragma("GCC diagnostic ignored \"-Wformat-security\"")
#define LKT_IGNORE_WARN_WRITE_STRING _Pragma("GCC diagnostic ignored \"-Wwrite-strings\"")
#define LKT_POP                      _Pragma("GCC diagnostic pop")

#elif defined(_MSC_VER)
#error "MS compiler not supported"

#else
#warn "Unknown compiler"
#define LKT_PUSH
#define LKT_IGNORE_WARN_FMT_SECURITY
#define LKT_POP

#endif

/* Set the name of a thread, OS specific */
void lkt_thread_set_name(const char *name);

/* Get the path to the currently executing binary */
int lkt_read_self_exe(char *path, size_t len);

/* OS specific defines */
#define SELF_EXECUTABLE_LINUX   "/proc/self/exe"
#define SELF_EXECUTABLE_FREEBSD "/proc/curproc/file"
#define SELF_EXECUTABLE_SOLARIS "/proc/self/path/a.out"

#if defined(__linux__)
#if !defined(WCONTINUED)
#define WCONTINUED 0x00000008
#endif
#endif

/* Plateform specific functions */

/* Same as `be_uint64_t` but for Big-endian doubles. The result is a
   correct endian number. Restriction: bytes must be of length 8.
   Use the be_uint64_t defined function. */
double be_double_t(const uint8_t bytes[]);

/* Same as `be_double_t` but for floats. Restriction: bytes must
   be of length 4. The result is a correct endian number Use the
   be_uint32_t defined function. */
float be_float_t(const uint8_t bytes[]);

/* Read `bytes` as the big endian representation of a 32-bit unsigned integer.
   `n` is the number of bytes in `bytes`. The result is a 'correct'-endian.
   Returns 0 if n is 0.
   Restriction: n <= 4 */
uint32_t be_uint32_t(const uint8_t bytes[], size_t n);

/* Same as `be_uint32_t` but for 64-bit unsigned integers. The result is a
   'correct'-endian. Restriction: n <= 8 */
uint64_t be_uint64_t(const uint8_t bytes[], size_t n);

#if defined(__cplusplus)
}
#endif
#endif /* LKT_OS_H__ */
