#if !defined(LKT_QUEUE_H__)
#define LKT_QUEUE_H__

#if defined(__cplusplus)
extern "C" {
#endif

#if !defined(_POSIX_C_SOURCE)
#define _POSIX_C_SOURCE 200809L
#endif

/* Everything in this file is implemented in the header, so that modules don't
   have to compile the lib with another .c file. */

#include <lektor/common.h>

enum {
    __LKT_PLAY_STOP   = 0,
    __LKT_PLAY_PLAY   = 1,
    __LKT_PLAY_PAUSE  = 2,
    __LKT_PLAY_TOGGLE = 3,
};

enum {
    __LKT_DB_UPDATING_FINISHED = 0, /* Not updating       */
    __LKT_DB_UPDATING_PROGRESS = 1, /* Update in progress */
};

#define LKT_PLAY_STOP   ((void *)(size_t)__LKT_PLAY_STOP)
#define LKT_PLAY_PLAY   ((void *)(size_t)__LKT_PLAY_PLAY)
#define LKT_PLAY_PAUSE  ((void *)(size_t)__LKT_PLAY_PAUSE)
#define LKT_PLAY_TOGGLE ((void *)(size_t)__LKT_PLAY_TOGGLE)

#define LKT_DB_UPDATING_PROGRESS ((void *)(size_t)__LKT_DB_UPDATING_PROGRESS)
#define LKT_DB_UPDATING_FINISHED ((void *)(size_t)__LKT_DB_UPDATING_FINISHED)

typedef enum {
    LKT_EVENT_NULL            = 0,         // NULL
    LKT_EVENT_PLAY_POS        = (1 << 1),  // size_t
    LKT_EVENT_PLAY_FILE       = (1 << 2),  // XXX: UNUSED
    LKT_EVENT_PLAY_NEXT       = (1 << 3),  // NULL
    LKT_EVENT_PLAY_PREV       = (1 << 4),  // NULL
    LKT_EVENT_PLAY_TOGGLE     = (1 << 5),  // size_t, `LKT_PLAY_.*`
    LKT_EVENT_PROP_VOL        = (1 << 6),  // size_t
    LKT_EVENT_PROP_DUR        = (1 << 7),  // size_t
    LKT_EVENT_PROP_TIME       = (1 << 8),  // size_t
    LKT_EVENT_SKIP_CURRENT    = (1 << 9),  // NULL
    LKT_EVENT_DB_UPDATING     = (1 << 10), // size_t, `LKT_DB_UPDATING_*`
    LKT_EVENT_DB_UPDATE_TOTAL = (1 << 11), // size_t, increments udapte_total
    LKT_EVENT_DB_UPDATE_TICK  = (1 << 12), // NULL, increment the `update_current`
    LKT_EVENT_TOUCH_KARA      = (1 << 13), // size_t, kara id
} LKT_EVENT_TYPE;

#define LKT_EVENT_PLAY                                                                      \
    (LKT_EVENT_PLAY_POS | LKT_EVENT_PLAY_FILE | LKT_EVENT_PLAY_NEXT | LKT_EVENT_PLAY_PREV | \
     LKT_EVENT_PLAY_TOGGLE | LKT_EVENT_SKIP_CURRENT)
#define LKT_EVENT_PROP (LKT_EVENT_PROP_VOL | LKT_EVENT_PROP_DUR | LKT_EVENT_PROP_TIME)
#define LKT_EVENT_UPDATE                                                            \
    (LKT_EVENT_DB_UPDATING | LKT_EVENT_DB_UPDATE_TICK | LKT_EVENT_DB_UPDATE_TOTAL | \
     LKT_EVENT_TOUCH_KARA)

typedef struct {
    unsigned int type;
    void *attr;
} lkt_event;

struct queue {
    pthread_mutex_t lock;

    volatile lkt_event *volatile contents;
    volatile size_t size;
    volatile size_t last;

    volatile int available;
};

/* Initialize the queue. Can't fail if the passed argument is not NULL.
 * WARN: In case of out of memory this function will abort the process. */
PRIVATE_FUNCTION int
lkt_queue_new(struct queue *ret)
{
    if (!ret)
        return 1;

    pthread_mutex_t mxt = PTHREAD_MUTEX_INITIALIZER;
    struct queue _ret   = {
        .contents  = LKT_ALLOC_ARRAY(lkt_event, LKT_DEFAULT_LIST_SIZE),
        .size      = LKT_DEFAULT_LIST_SIZE,
        .last      = 0,
        .available = 0,
        .lock      = mxt,
    };

    *ret = _ret;
    return 0;
}

/* Free the queue. Can't fail. */
PRIVATE_FUNCTION void
lkt_queue_free(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    if (queue && queue->contents)
        free((void *)queue->contents);
    pthread_mutex_unlock(&queue->lock);
}

/* Is an event of the following type present in the queue?
 * Returns 0 if the type is not present, 1 otherwise. */
PRIVATE_FUNCTION bool
lkt_queue_has_event(struct queue *queue, LKT_EVENT_TYPE type)
{
    bool ret = false;
    pthread_mutex_lock(&queue->lock);
    for (size_t i = 0; i < queue->last; ++i) {
        if (queue->contents[i].type & type) {
            ret = true;
            break;
        }
    }
    pthread_mutex_unlock(&queue->lock);
    return ret;
}

/* Send an event into the queue */
PRIVATE_FUNCTION void
lkt_queue_send(struct queue *queue, LKT_EVENT_TYPE _type, void *_attr)
{
    pthread_mutex_lock(&queue->lock);
    GOTO_UNLESS(queue, "Invalid argument", end)

    volatile lkt_event *new;
    if (queue->size == queue->last) {
        new = realloc((void *)queue->contents, queue->size * 2 * sizeof(lkt_event));

        if (NULL == new)
            goto end;

        queue->contents = new;
        queue->size *= 2;
    }

    lkt_event evt = {
        .type = _type,
        .attr = _attr,
    };
    queue->contents[(queue->last)++] = evt;

    if (!(_type & queue->available))
        LOG_WARN("QUEUE", "The event %d is not available, push it anyway", _type);
end:
    pthread_mutex_unlock(&queue->lock);
}

PRIVATE_FUNCTION lkt_event
lkt_queue_handle(struct queue *queue)
{
    pthread_mutex_lock(&queue->lock);
    lkt_event ret = { 0 };
    if (!queue || !queue->last)
        goto end;
    if (!(queue->contents[0].type & queue->available) && queue->contents[0].type) {
        LOG_WARN("QUEUE", "Event %d is not available", queue->contents[0].type);
        goto end;
    }

    ret = queue->contents[0];
    for (size_t i = 1; i < queue->last; ++i)
        queue->contents[i - 1] = queue->contents[i];
    queue->last -= 1;
end:
    pthread_mutex_unlock(&queue->lock);
    return ret;
}

PRIVATE_FUNCTION void
lkt_queue_make_available(struct queue *queue, LKT_EVENT_TYPE type)
{
    pthread_mutex_lock(&queue->lock);
    queue->available |= type;
    pthread_mutex_unlock(&queue->lock);
}

#if defined(__cplusplus)
}
#endif
#endif /* LKT_QUEUE_H__ */
