#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/config.h>
#include <lektor/database.h>
#include <lektor/net.h>
#include <lektor/reg.h>

#include <dlfcn.h>
#include <pwd.h>

static inline char *
strip(char *s)
{
    char *p = s + strlen(s);
    while (p > s && isspace(*(--p)))
        *p = 0;
    return s;
}

static inline char *
skip(char *s)
{
    while (*s && isspace(*s))
        s++;
    return s;
}

static int
handler(lkt_db *user, const char *section, const char *name, const char *value)
{
    RETURN_UNLESS(section && name && value, "I can't complete the database with incomplete lines",
                  1);
    RETURN_UNLESS(database_config_set(user, section, name, value), "Failed to update the database",
                  1);
    return 0;
}

static inline void
__set_log_level(const char *name, const char *level)
{
    if (!STR_MATCH(name, "log")) {
        LOG_WARN("CONFIG", "Invalid option '%s[:=]%s' with no section", name, level);
        return;
    }

    if (!level[0]) {
        LOG_WARN("CONFIG", "Invalid empty 'log' option");
        return;
    }

    if (STR_MATCH(level, "error"))
        lkt_set_log_level(LOG_LEVEL_ERROR);
    else if (STR_MATCH(level, "warn") || STR_MATCH(level, "warning"))
        lkt_set_log_level(LOG_LEVEL_WARN);
    else if (STR_MATCH(level, "info"))
        lkt_set_log_level(LOG_LEVEL_INFO);
    else if (STR_MATCH(level, "debug"))
        lkt_set_log_level(LOG_LEVEL_DEBUG);
    else
        lkt_set_log_level(strtol(level, NULL, 0));

    LOG_INFO("CONFIG", "Log level set to %d", lkt_get_log_level());
}

static inline int
ini_parse(const char *path, lkt_db *db)
{
    char *start, *end, *name, *value;
    char section[LKT_LINE_MAX], line[LKT_LINE_MAX];
    int error = 0, linenum = 0, len;
    FILE *file = fopen(path, "r");
    if (!file) {
        LOG_ERROR("PARSER", "Failed to open config file '%s'", path);
        return 1;
    }

    memset(section, 0, LKT_LINE_MAX);
    memset(line, 0, LKT_LINE_MAX);

    LOG_INFO("CONFIG", "Save config to database from file");

    /* Parse the file */
    while (NULL != fgets(line, LKT_LINE_MAX, file)) {
        ++linenum;
        start = skip(strip(line));

        /* Skip comments */
        if (strspn(start, ";#"))
            continue;

        /* Handle sections */
        else if (start[0] == '[') {
            end = &start[1 + strcspn(start + 1, "]")];
            if (end[0] == ']') {
                end[0] = '\0';
                len    = strlen(&start[1]);
                len    = MIN(len + 1, LKT_LINE_MAX);
                memcpy(section, &start[1], len);
                section[LKT_LINE_MAX - 1] = '\0';
            }

            else {
                error = 1;
                LOG_ERROR("PARSER", "Invalid section name at line '%d'", linenum);
            }
        }

        /* Handle name[:=]name pair */
        else if (start[0]) {
            end = &start[1 + strcspn(start + 1, ":=")];
            if (end[0] == '=' || end[0] == ':') {
                end[0] = '\0';
                name   = strip(start);
                value  = &end[1];

                /* Find a comment */
                end = &value[strcspn(value, ";#")];
                if (end[0])
                    end[0] = '\0';

                /* Skip all spaces */
                value = skip(value);
                strip(value);

                /* Handle the SECTION, NAME[:=]VALUE
                   The only option that has no SECTION is the log level:
                   log[:=]ERROR|WARN|INFO|DEBUG|\d+ */
                if (section[0]) {
                    if (handler(db, section, name, value)) {
                        error = 1;
                        LOG_ERROR("PARSER", "Failed to '[handle] %s, %s{:,=}%s' at line '%d'",
                                  section, name, value, linenum);
                    }
                } else
                    __set_log_level(name, value);
            }

            else {
                error = 1;
                LOG_ERROR("PARSER", "Invalid name[:=]value pair at line '%d'", linenum);
            }
        }
    }

    /* End of the function */
    fclose(file);
    if (error)
        LOG_ERROR("PARSER", "An error occured while parsing the file '%s'", path);
    return error;
}

void
config_default_file(char *dest, size_t len)
{
#pragma message( \
    TODO         \
    "Use ~/.config/baka instead of ~/.config/lektor for all applications (lektor, amadeus, etc)")
    /* First try the XDG_CONFIG_HOME variable, else the default location HOME/.config. */
    memset(dest, 0, len * sizeof(char));
    char *home = getenv("XDG_CONFIG_HOME");
    if (NULL == home || strlen(home) >= len) {
        LOG_DEBUG("CONFIG", "No env variable XDG_CONFIG_HOME");
        home = getenv("HOME");
        if (NULL == home)
            LOG_FATAL("Failed to get home folder for user, will now exit");
        safe_strncpy(dest, home, len);
        strncat(dest, "/.config/lektor/lektor.ini", len - 1);
        LOG_DEBUG("CONFIG", "Use '%s' insted of XDG_CONFIG_HOME", home);
    } else
        strncat(dest, "/lektor/lektor.ini", len - 1);
}

int
config_detect_file(char *conf, size_t conf_len)
{
    bool is_malloc    = false;
    struct passwd *pw = getpwuid(getuid());
    char *home;

    if (conf == NULL)
        return 1;

    memset(conf, 0, conf_len * sizeof(char));

    /* Try the current working dir config file. */
    if (getcwd(conf, conf_len - 1) != NULL) {
        strncat(conf, "/lektor.ini", conf_len - 1);
        LOG_INFO("CONFIG", "Trying %s", conf);
        if (!access(conf, R_OK | F_OK))
            goto found;
    }

    /* Try the config file from the config directory. */
    home = getenv("XDG_CONFIG_HOME");
    if (home && strlen(home) < conf_len) {
        /* Skip the strncat to not append the '.config/' to
           the XDG_CONFIG_HOME which must already have this directory in it */
        memcpy(conf, home, (strlen(home) + 1) * sizeof(char));
        strncat(conf, "/lektor/lektor.ini", conf_len - 1);
        goto skip_this_strcat;
    }
    if (!home || (strlen(home) >= conf_len))
        home = getenv("HOME");
    if (!home || (strlen(home) >= conf_len))
        home = pw->pw_dir;
    if (!home || (strlen(home) >= conf_len))
        goto no_config_directory;
    memcpy(conf, home, (strlen(home) + 1) * sizeof(char));
    strncat(conf, "/.config/lektor/lektor.ini", conf_len - 1);
skip_this_strcat:
    LOG_INFO("CONFIG", "Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

no_config_directory:
    /* Try the '/opt/lektor' file. */
    memcpy(conf, "/opt/lektor/lektor.ini", sizeof("/opt/lektor/lektor.ini"));
    LOG_INFO("CONFIG", "Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

    /* Try the '/usr/local/etc/lektor.ini' file. */
    memcpy(conf, "/usr/local/etc/lektor.ini", sizeof("/usr/local/etc/lektor.ini"));
    LOG_INFO("CONFIG", "Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

    /* Try the '/etc/lektor.ini' file. */
    memcpy(conf, "/etc/lektor.ini", sizeof("/etc/lektor.ini"));
    LOG_INFO("CONFIG", "Trying %s", conf);
    if (!access(conf, R_OK | F_OK))
        goto found;

    /* Error */
    LOG_ERROR("CONFIG", "An error occured with file %s: %s", conf, strerror(errno));
    if (is_malloc)
        free(conf);
    return 1;
found:
    LOG_INFO("CONFIG", "Using file %s", conf);
    return 0;
}

int
config_new(lkt_db *db, const char *conf)
{
    if (ini_parse(conf, db)) {
        LOG_ERROR("CONFIG", "Failed to parse file %s", conf);
        goto error;
    }

    if (!database_validate_conf(db)) {
        LOG_ERROR("CONFIG", "Configuration file %s is incomplete", conf);
        goto error;
    }

    return 0;
error:
    LOG_ERROR("CONFIG", "Errors detected, here is a default config");
    config_default(stdout);
    return 1;
}

void
config_default(FILE *output)
{
    fwrite(lkt_default_config_file, sizeof(char), strlen(lkt_default_config_file), output);
}

int
config_open(lkt_db *db)
{
    char conf_file[PATH_MAX];
    RETURN_IF(config_detect_file(conf_file, PATH_MAX), "Error while searching for a config file",
              1);
    RETURN_IF(config_new(db, conf_file), "Failed to read configuration file", 1);
    return 0;
}

void
config_handle_hook(lkt_db *db, const char *hook_name)
{
    char hook_handlers[LKT_LINE_MAX];
    if (!database_config_get_text_nospace(db, "hook", hook_name, hook_handlers, LKT_LINE_MAX)) {
        LOG_ERROR("HOOK", "Failed to get the hook handlers list for hook '%s'", hook_name);
        return;
    }

    size_t save = 0;
    char name[LKT_LINE_MAX];
    int code;
    struct module_reg *server_reg;
    reg_global(&server_reg);

    while (0 == (code = iter_string(hook_handlers, ",", name, LKT_LINE_MAX, &save))) {
        if (STR_MATCH("none", name))
            continue;
        reg_call(server_reg, name, 1, db);
    }

    if (code == -1)
        LOG_ERROR("HOOK", "Failed to parse the handle list for hook '%s'", hook_name);
}
