#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/database.h>
#include <lektor/internal/dbmacro.h>
#include <math.h>

/* Forward declarations for set operations */
PRIVATE_FUNCTION bool database_set_kara_mtime_id(lkt_db *, int, uint64_t);
PRIVATE_FUNCTION bool database_set_kara_duration_id(lkt_db *, int, uint64_t);

/* Is cached? */
PRIVATE_FUNCTION bool
database_is_kara_cached(lkt_db *db, int id)
{
    static const char *SQL = "SELECT cached_mtime, cached_duration FROM kara WHERE id = ?;";
    sqlite3_stmt *stmt     = NULL;
    bool ret_code          = false;

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);

    ret_code = ((0 != sqlite3_column_int(stmt, 0)) && (0 != sqlite3_column_int(stmt, 1)));

error:
    sqlite3_finalize(stmt);
    return ret_code;
}

PRIVATE_FUNCTION long
get_mtime(const char *path)
{
    struct stat statbuf = { 0 };
    return (stat(path, &statbuf) == -1) ? 0 : statbuf.st_mtime;
}

PRIVATE_FUNCTION void
___database_cache_kara_magic_by_file(lkt_db *db, const char *filename)
{
    /* Update the mkv magic error */
    int magic_error        = mkv_check_magic(filename);
    sqlite3_stmt *stmt     = NULL;
    static const char *SQL = "UPDATE kara_cache SET magic_error = ? WHERE file_path = ?;";
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, magic_error, error);
    SQLITE_BIND_TEXT(db, stmt, 2, filename, error);
    SQLITE_STEP_OK(db, stmt, error);
    sqlite3_finalize(stmt);
    return;

error:
    LOG_ERROR("CACHE", "Failed to set mkv magic error to %d for kara %s", magic_error, filename);
    sqlite3_finalize(stmt);
    return;
}

void
database_cache_kara_magic(lkt_db *db, int id)
{
    /* Get the filepath */
    char filename[PATH_MAX];
    memset(filename, 0, sizeof(filename));
    database_get_kara_path(db, id, filename);
    if (filename[0] == 0) {
        LOG_ERROR("CACHE", "Failed to get filename for kara %ld", id);
        return;
    }

    /* Update the mkv magic error */
    ___database_cache_kara_magic_by_file(db, filename);
}

bool
database_is_kara_mkv(lkt_db *db, char filepath[PATH_MAX], bool *valid_mkv)
{
    static const char *SQL = "SELECT magic_error FROM kara_cache WHERE file_path = ?;";
    sqlite3_stmt *stmt     = NULL;
    int ret_code           = false;
    int is_valid_mkv       = 0;
    bool try_once          = false;
    *valid_mkv             = false;

    /* About the magic_error:
     * - 0: OK
     * - 1: ACK
     * - 2: Not checked */

retry:
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_TEXT(db, stmt, 1, filepath, error);
    SQLITE_STEP_ROW(db, stmt, error);
    is_valid_mkv = sqlite3_column_int(stmt, 0);
    if ((is_valid_mkv == 2) && (!try_once)) {
        ___database_cache_kara_magic_by_file(db, filepath);
        try_once = true;
        goto retry;
    }

    *valid_mkv = (is_valid_mkv == 0);
    ret_code   = true;
error:
    sqlite3_finalize(stmt);
    return ret_code;
}

/* Request to cache a kara */
void
database_cache_kara(lkt_db *db, int id)
{
    char filepath[PATH_MAX];

    if (!database_get_kara_path(db, id, filepath)) {
        LOG_ERROR("CACHE", "Failed to get a path for kara %d, don't update cache", id);
        return;
    }

    if (database_is_kara_cached(db, id)) {
        LOG_INFO("CACHE", "Kara %d is already cached", id);
        return;
    } else
        LOG_INFO("CACHE", "Trying to cache kara %d", id);

    long mtime = get_mtime(filepath);
    if (!database_set_kara_mtime_id(db, id, mtime))
        LOG_ERROR("CACHE", "Failed to cache mtime for '%s' kara %d to %ld", filepath, id, mtime);

    double duration = 0.;
    if (kara_read_length(&duration, filepath) || isnan(duration))
        LOG_ERROR("CACHE", "Failed to read lengh of '%s' for kara %d", filepath, id);

    uint64_t duration_uint = (uint64_t)duration;
    if (!database_set_kara_duration_id(db, id, duration_uint))
        LOG_ERROR("CACHE", "Failed to cache duration for '%s' kara %d to %ld", filepath, id,
                  duration_uint);

    LOG_INFO("CACHE", "Cached mtime %ld and duration %ld for '%s' kara %d", mtime, duration_uint,
             filepath, id);
}

#define ___cache_set_from_id(what)                                                                 \
    PRIVATE_FUNCTION bool database_set_kara_##what##_id(lkt_db *db, int id, uint64_t what)         \
    {                                                                                              \
        bool ret_code      = false;                                                                \
        sqlite3_stmt *stmt = NULL;                                                                 \
        /* Update the disk cache */                                                                \
        SQLITE_PREPARE(db, stmt, "UPDATE kara SET cached_" #what " = ? WHERE id = ?;", error);     \
        SQLITE_BIND_INT(db, stmt, 1, what, error);                                                 \
        SQLITE_BIND_INT(db, stmt, 2, id, error);                                                   \
        SQLITE_STEP_OK(db, stmt, error);                                                           \
        /* Update the memory cache */                                                              \
        SQLITE_PREPARE(db, stmt, "UPDATE kara_cache SET " #what " = ? WHERE kara_id = ?;", error); \
        SQLITE_BIND_INT(db, stmt, 1, what, error);                                                 \
        SQLITE_BIND_INT(db, stmt, 2, id, error);                                                   \
        SQLITE_STEP_OK(db, stmt, error);                                                           \
        ret_code = true;                                                                           \
    error:                                                                                         \
        sqlite3_finalize(stmt);                                                                    \
        return ret_code;                                                                           \
    }

#define ___cache_get_from_id(what)                                                              \
    bool database_get_kara_##what##_id(lkt_db *db, int id, uint64_t *what)                      \
    {                                                                                           \
        bool ret_code      = false;                                                             \
        bool retry_once    = false;                                                             \
        sqlite3_stmt *stmt = NULL;                                                              \
    retry:                                                                                      \
        *what = 0;                                                                              \
        SQLITE_PREPARE(db, stmt, "SELECT " #what " FROM kara_cache WHERE kara_id = ?;", error); \
        SQLITE_BIND_INT(db, stmt, 1, id, error);                                                \
        SQLITE_STEP_ROW(db, stmt, error);                                                       \
        *what = sqlite3_column_int64(stmt, 0);                                                  \
        if ((*what) == 0 && !retry_once) {                                                      \
            database_cache_kara(db, id);                                                        \
            retry_once = true;                                                                  \
            goto retry;                                                                         \
        }                                                                                       \
        ret_code = true;                                                                        \
    error:                                                                                      \
        sqlite3_finalize(stmt);                                                                 \
        return ret_code;                                                                        \
    }

/* Define get operation on the cache */
#define ___cache_get_from_path(what)                                                          \
    bool database_get_kara_##what##_path(lkt_db *db, char filepath[PATH_MAX], uint64_t *what) \
    {                                                                                         \
        int id = 0;                                                                           \
        if (!database_get_kara_id(db, filepath, &id)) {                                       \
            LOG_ERROR("CACHE", "Can't find kara with path '%s'", filepath);                   \
            return false;                                                                     \
        }                                                                                     \
        return database_get_kara_##what##_id(db, id, what);                                   \
    }

___cache_set_from_id(mtime);
___cache_set_from_id(duration);
___cache_get_from_id(mtime);
___cache_get_from_id(duration);
___cache_get_from_path(mtime);
___cache_get_from_path(duration);
