#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/database.h>
#include <lektor/net.h>
#include <lektor/internal/dbmacro.h>

PRIVATE_FUNCTION int
__check_sticker_type(const char *type)
{
    return !(STR_MATCH(type, "kara") || STR_MATCH(type, "plt"));
}

/* Private search implementation. */
struct lkt_search {
    lkt_db *db;
    sqlite3_stmt *stmt;
    enum lkt_search_type {
        lkt_search_database,
        lkt_search_playlist,
        lkt_search_queue,
        lkt_search_sticker,
        lkt_search_listplaylist,
    } type;

    void (*call)(void); /* Called during the iter phase, casted     */

    struct lkt_state *srv;
    size_t c;

    long continuation; /* The continuation state of the client      */
    int msg_count;     /* How much messages we can send             */

    union {
        struct lkt_sticker_opt st; /* Sticker construct structure   */
        const char *plt_name;      /* Playlist name                 */
    };

    struct lkt_uri ka_uri; /* Search by uri                         */
};

/* Getter functions */

struct lkt_state *
database_search_get_srv(struct lkt_search *search)
{
    return search->srv;
}

size_t
database_search_get_client(struct lkt_search *search)
{
    return search->c;
}

size_t
database_search_get_continuation(struct lkt_search *search)
{
    return search->continuation;
}

/* Free function */
void
database_search_free(struct lkt_search *search)
{
    if (search == NULL)
        return;

    if (search->stmt != NULL)
        sqlite3_finalize(search->stmt);
    free(search);
}

/* Allocation function */
struct lkt_search *
database_search_new(struct lkt_state *srv, size_t c, int cont, function_ptr call)
{
    /* Minus 3 for control messages. If no server do it `X` by `X`. */
    size_t msg_count = srv != NULL ? lkt_remaining_msg(srv, c) - 3 : LKT_DB_DEFAULT_ITERATION_MAX;
    struct lkt_search *ret_ptr = LKT_ALLOC_STRUCT(lkt_search);
    struct lkt_search ret      = {
        .srv          = srv,
        .c            = c,
        .continuation = cont,
        .msg_count    = msg_count,
        .call         = call,
        .stmt         = NULL,
    };

    *ret_ptr = ret;
    return ret_ptr;
}

/* Copy function */
struct lkt_search *
database_search_new_from(struct lkt_search *from)
{
    struct lkt_search *ret_ptr = LKT_ALLOC_STRUCT(lkt_search);
    memcpy(ret_ptr, from, sizeof(struct lkt_search));
    ret_ptr->stmt = NULL; /* Omit that one, the returned search is not initialized. */
    return ret_ptr;
}

/* Setter functions */

void
database_search_set_continuation(struct lkt_search *search, size_t cont)
{
    search->continuation = cont;
}

void
database_search_set_sticker(struct lkt_search *search, struct lkt_sticker_opt *sticker)
{
    search->st = *sticker;
}

void
database_search_set_uri(struct lkt_search *search, struct lkt_uri *ka_uri)
{
    search->ka_uri = *ka_uri;
}

void
database_search_set_name(struct lkt_search *search, const char *pltname)
{
    search->plt_name = pltname;
}

bool
database_search_database_init(lkt_db *db, struct lkt_search *ret)
{
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    static const char *SQL_STMT_TEMPLATE =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
        " FROM kara WHERE %s LIKE ?)"
        "SELECT id, string, (SELECT MAX(len) FROM content)"
        "FROM content LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_database;

    /* Search part */
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE, ret->ka_uri.column_name,
                  ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    if (!ret->ka_uri.is_int) {
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->ka_uri.value, error);
        LOG_DEBUG("DB", "Uri was a text");
    } else {
        SQLITE_BIND_INT(db, ret->stmt, 1, (int)ret->ka_uri.id, error);
        LOG_DEBUG("DB", "Uri was a size_t");
    }
    ret->db = db;
    return true;

error:
    database_search_free(ret);
    return false;
}

bool
database_search_playlist_init(lkt_db *db, struct lkt_search *ret)
{
    /* Sql queries definitions */
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    static const char *SQL_TEMPLATE =
        "WITH content AS ("
        " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
        " FROM kara_playlist"
        " JOIN playlist ON playlist.id = playlist_id"
        "               AND playlist.name COLLATE nocase = ?"
        " JOIN kara ON kara.id = kara_id"
        "           AND %s LIKE ?) "
        "SELECT id, string, (SELECT MAX(len) FROM content) "
        "FROM content LIMIT %d OFFSET %d;";
    static const char *SQL = "WITH content AS ("
                             " SELECT kara.id AS id, string, LENGTH(CAST(kara.id AS TEXT)) AS len"
                             " FROM kara_playlist"
                             " JOIN playlist ON playlist.id = playlist_id"
                             "               AND playlist.name COLLATE NOCASE = ?"
                             " JOIN kara ON kara.id = kara_id) "
                             "SELECT id, string, (SELECT MAX(len) FROM content) "
                             "FROM content LIMIT %d OFFSET %d;";

    /* Common part */
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    ret->type = lkt_search_playlist;
    ret->db   = db;
    if (ret->ka_uri.type == URI_NULL)
        goto uri_is_null;

    /* Filter with an uri */
    LOG_INFO("DB", "List plt '%s' with non null uri", ret->plt_name);
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_TEMPLATE, ret->ka_uri.column_name,
                  ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->plt_name, error);

    if (!ret->ka_uri.is_int) {
        SQLITE_BIND_TEXT(db, ret->stmt, 2, ret->ka_uri.value, error);
        LOG_DEBUG("DB", "Uri was a text");
    }

    else {
        SQLITE_BIND_INT(db, ret->stmt, 2, (int)ret->ka_uri.id, error);
        LOG_DEBUG("DB", "Uri was a size_t");
    }

    return true;

    /* Don't filter */
uri_is_null:
    LOG_INFO("DB", "List plt '%s' with null uri", ret->plt_name);
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL, ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->plt_name, error);
    return true;

    /* Error */
error:
    database_search_free(ret);
    return false;
}

bool
database_search_listplaylist_init(lkt_db *db, struct lkt_search *ret)
{
    static const char *SQL = "SELECT name FROM playlist LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL, ret->msg_count, ret->continuation);
    ret->type = lkt_search_listplaylist;
    ret->db   = db;
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);
    LOG_INFO("DB", "List playlists with LIMIT %d and OFFSET %d", ret->msg_count,
             (int)ret->continuation);
    return true;

error:
    database_search_free(ret);
    return false;
}

bool
database_search_sticker_init(lkt_db *db, struct lkt_search *ret)
{
    if (__check_sticker_type(ret->st.type)) {
        LOG_ERROR("DB", "Type '%s' is invalid", ret->st.type);
        return false;
    }

    static const char *SQL_all_types = "SELECT name, sts.id, value FROM 'stickers' JOIN "
                                       "( SELECT id, sticker, value, 'kara' FROM 'stickers.kara'"
                                       "  UNION"
                                       "  SELECT id, sticker, value , 'plt' FROM 'stickers.plt'"
                                       ") AS sts"
                                       "ON sts.sticker = 'stickers'.id";
    static const char *SQL_one_type = "SELECT name, sts.id, value, '%s' "
                                      "FROM 'stickers.%s' AS sts "
                                      "JOIN 'stickers' "
                                      "ON 'stickers'.id = sts.sticker";
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_sticker;

    if (ret->st.type == NULL)
        memcpy(SQL, SQL_all_types, strlen(SQL_all_types) + 1);
    else
        sprintf(SQL, SQL_one_type, ret->st.type, ret->st.type);
    if (ret->st.uri != 0)
        sprintf(SQL + strlen(SQL), " AND sts.id = %d", ret->st.uri);
    if (ret->st.op != 0)
        sprintf(SQL + strlen(SQL), " AND sts.value %s %d",
                ret->st.op == '>'   ? ">="
                : ret->st.op == '<' ? "<="
                                    : "=",
                ret->st.value);
    strcat(SQL, !ret->st.name ? ";" : " AND name = ?;");

    SQLITE_PREPARE(db, ret->stmt, SQL, error);
    if (ret->st.name)
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->st.name, error);
    return true;

error:
    database_search_free(ret);
    return false;
}

bool
database_search_queue_init(lkt_db *db, struct lkt_search *ret)
{
    RETURN_UNLESS(ret, "Exit because return pointer is NULL", false);
    static const char *SQL_STMT_TEMPLATE = "WITH content AS ("
                                           " SELECT"
                                           "  kara_id AS id,"
                                           "  string,"
                                           "  LENGTH(CAST(kara_id AS TEXT)) AS len"
                                           " FROM queue"
                                           " JOIN kara "
                                           " ON kara_id = kara.id AND %s LIKE ?)"
                                           "SELECT id, string, (SELECT MAX(len) FROM content)"
                                           "FROM content LIMIT %d OFFSET %d;";
    char SQL_STMT[LKT_MAX_SQLITE_STATEMENT];
    ret->type = lkt_search_queue;

    /* Search part */
    safe_snprintf(SQL_STMT, LKT_MAX_SQLITE_STATEMENT, SQL_STMT_TEMPLATE, ret->ka_uri.column_name,
                  ret->msg_count, ret->continuation);
    SQLITE_PREPARE(db, ret->stmt, SQL_STMT, error);

    if (!ret->ka_uri.is_int) {
        SQLITE_BIND_TEXT(db, ret->stmt, 1, ret->ka_uri.value, error);
        LOG_DEBUG("DB", "Uri was a text");
    }

    else {
        SQLITE_BIND_INT(db, ret->stmt, 1, (int)ret->ka_uri.id, error);
        LOG_DEBUG("DB", "Uri was a size_t");
    }

    ret->db = db;
    return true;

error:
    database_search_free(ret);
    return false;
}

bool
database_search_iter(struct lkt_search *item)
{
    const char *sql_row, *type;
    int id, code, id_len;
    RETURN_UNLESS(item, "Exit because item is NULL, end iterations", false);
    GOTO_UNLESS(item->call, "NULL call function, terminate the search", end);
    code = sqlite3_step(item->stmt);

    if (code == SQLITE_DONE)
        goto end;

    if (code != SQLITE_ROW) {
        LOG_ERROR("DB", "Step failed, expected a ROW or a DONE: %s",
                  sqlite3_errmsg((sqlite3 *)item->db));
        goto end;
    }

    switch (item->type) {
    case lkt_search_database:
    case lkt_search_queue:
    case lkt_search_playlist:
        id      = sqlite3_column_int(item->stmt, 0);
        sql_row = sqlite3_column_chars(item->stmt, 1);
        id_len  = sqlite3_column_int(item->stmt, 2);
        return ((lkt_search_database_func)item->call)(item->srv, item->c, id, id_len, sql_row);

    case lkt_search_sticker:
        sql_row = sqlite3_column_chars(item->stmt, 0); /* Name */
        id      = sqlite3_column_int(item->stmt, 1);   /* Id   */
        code    = sqlite3_column_int(item->stmt, 2);   /* Val  */
        type    = sqlite3_column_chars(item->stmt, 3); /* Type */
        return ((lkt_search_sticker_func)item->call)(item->srv, item->c, sql_row, type, id, code);

    case lkt_search_listplaylist:
        sql_row = sqlite3_column_chars(item->stmt, 0); /* Name */
        return ((lkt_search_listplaylist_func)item->call)(item->srv, item->c, sql_row);

    default: LOG_WARN("DB", "Search type %d is not implemented", item->type); goto end;
    }

end:
    database_search_free(item);
    return false;
}

bool
database_kara_by_id(lkt_db *db, int id, struct kara_metadata *res, char filepath[PATH_MAX])
{
    static const char *SQL = "SELECT song_name, source_name, category, language, author_name, "
                             "song_type, song_number, file_path FROM kara WHERE id = ?;";
    sqlite3_stmt *stmt = 0;
    int ret            = false;

    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_BIND_INT(db, stmt, 1, id, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (!res)
        goto no_metadata;
    strncpy(res->song_name, sqlite3_column_chars(stmt, 0), LEKTOR_TAG_MAX - 1);
    strncpy(res->source_name, sqlite3_column_chars(stmt, 1), LEKTOR_TAG_MAX - 1);
    strncpy(res->category, sqlite3_column_chars(stmt, 2), LEKTOR_TAG_MAX - 1);
    strncpy(res->language, sqlite3_column_chars(stmt, 3), LEKTOR_TAG_MAX - 1);
    strncpy(res->author_name, sqlite3_column_chars(stmt, 4), LEKTOR_TAG_MAX - 1);
    strncpy(res->song_type, sqlite3_column_chars(stmt, 5), LEKTOR_TAG_MAX - 1);
    res->song_number = sqlite3_column_int(stmt, 6);
no_metadata:
    strncpy(filepath, sqlite3_column_chars(stmt, 7), PATH_MAX - 1);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

PRIVATE_FUNCTION void
___get_all_kara(lkt_db *db, int **ret_ids, size_t *ret_ids_len, const char *SQL_DATA,
                const char *SQL_COUNT)
{
    sqlite3_stmt *stmt = NULL;
    size_t index       = 0;

    /* Get storage for all IDs */
    SQLITE_PREPARE(db, stmt, SQL_COUNT, error);
    SQLITE_STEP_ROW(db, stmt, error);
    *ret_ids_len = sqlite3_column_int64(stmt, 0);
    if (*ret_ids_len == 0) {
        LOG_WARN("DB", "Got no kara for the specific 'all' query: %s", SQL_COUNT);
        sqlite3_finalize(stmt);
        return;
    }
    sqlite3_finalize(stmt);
    stmt     = NULL;
    *ret_ids = LKT_ALLOC_ARRAY(int, *ret_ids_len);

    /* Get all the IDs */
    SQLITE_PREPARE(db, stmt, SQL_DATA, error);
    FOR_EVER
    {
        SQLITE_STEP_ROW(db, stmt, done);
        (*ret_ids)[index] = sqlite3_column_int(stmt, 0);
        index++;
    }

done:
    LOG_INFO("DB", "Done retrieving all the kara IDs, %ld kara retrieved", *ret_ids_len);
    sqlite3_finalize(stmt);
    return;

error:
    LOG_ERROR("DB", "An error occured during sql queries... oupsy");
    if (stmt != NULL)
        sqlite3_finalize(stmt);
    if ((*ret_ids) != NULL)
        free(*ret_ids);
    *ret_ids     = NULL;
    *ret_ids_len = 0;
    return;
}

void
database_get_all_non_cached_kara(lkt_db *db, int **ret_ids, size_t *ret_ids_len)
{
    static const char *SQL_DATA =
        "SELECT id FROM kara WHERE cached_mtime = 0 OR cached_duration = 0;";
    static const char *SQL_COUNT =
        "SELECT count(id) FROM kara WHERE cached_mtime = 0 OR cached_duration = 0;";
    ___get_all_kara(db, ret_ids, ret_ids_len, SQL_DATA, SQL_COUNT);
}

void
database_get_all_cached_kara(lkt_db *db, int **ret_ids, size_t *ret_ids_len)
{
    static const char *SQL_DATA  = "SELECT kara_id FROM kara_cache;";
    static const char *SQL_COUNT = "SELECT count(kara_id) FROM kara_cache;";
    ___get_all_kara(db, ret_ids, ret_ids_len, SQL_DATA, SQL_COUNT);
}
