#define _POSIX_C_SOURCE 200809L

#include <lektor/common.h>
#include <lektor/stack.h>
#include <lektor/database.h>
#include <lektor/internal/dbmacro.h>

extern unsigned char ___src_database_memory_sql[];
extern unsigned char ___src_database_disk_sql[];

static const char *__pragmas = "PRAGMA case_sensitive_like = false;\n"
                               "PRAGMA encoding            = 'UTF-8';\n"
                               "PRAGMA synchronous         = 0;\n"
                               "PRAGMA journal_mode        = off;\n";

#define HEAP_LIMIT_SOFT 100 * 1024 * 1024
#define HEAP_LIMIT_HARD 150 * 1024 * 1024

#if defined(LKT_NO_SQLITE3_HARD_HEAP_LIMIT64_FUNCTION)
sqlite3_int64
sqlite3_hard_heap_limit64(sqlite3_int64 UNUSED limit)
{
    LOG_ERROR("DB", "No sqlite3_hard_heap_limit64 function on your system");
    return 0;
}
#endif

struct named_db {
    lkt_db *db;
    const char *name;
};
static struct stack db_stack;
static int db_stack_init   = 0;
static pthread_mutex_t mtx = PTHREAD_MUTEX_INITIALIZER;

PRIVATE_FUNCTION void
___dec(lkt_db *db, const char *name)
{
    char SQL[LKT_MAX_SQLITE_STATEMENT];
    safe_snprintf(SQL, LKT_MAX_SQLITE_STATEMENT, "UPDATE %s.misc SET opened = 0;", name);
    SQLITE_EXEC(db, SQL, error);
    LOG_INFO("DB", "Dec open count on db '%s'", name);
    return;
error:
    LOG_ERROR("DB", "Faield to close database");
}

void
database_close_all(void)
{
    LOG_INFO("GARBAGE", "Closing all db");
    if (pthread_mutex_trylock(&mtx))
        return;
    struct named_db *item;
    while (!stack_empty(&db_stack)) {
        stack_pop(&db_stack, (void **)&item);
        ___dec(item->db, item->name);
        free((void *)item->name);
        free(item);
    }
    LOG_INFO("GARBAGE", "All db closed");
    stack_free(&db_stack);
}

PRIVATE_FUNCTION void
___inc(lkt_db *db, const char *name, bool check)
{
    SQLITE_EXEC(db, "UPDATE misc SET opened = (SELECT opened + 1 FROM misc);", error);

    if (db_stack_init == 0) {
        stack_new(&db_stack);
        db_stack_init = 1;
    }

    struct named_db *item = LKT_ALLOC_STRUCT(named_db);
    item->name            = strdup(name);
    item->db              = db;
    stack_push(&db_stack, item);
    return;
error:
    LOG_ERROR("DB", "Database already in use");
    if (check)
        exit(EXIT_FAILURE);
    ___dec(db, name);
}

bool
database_new(lkt_db **db)
{
    *db              = NULL;
    static int flags = SQLITE_OPEN_READWRITE | SQLITE_OPEN_SHAREDCACHE | SQLITE_OPEN_NOFOLLOW |
                       SQLITE_OPEN_FULLMUTEX;

    // clang-format off
    if (SQLITE_OK != sqlite3_enable_shared_cache(1)    ||
        sqlite3_soft_heap_limit64(HEAP_LIMIT_SOFT) < 0 ||
        sqlite3_hard_heap_limit64(HEAP_LIMIT_HARD) < 0 ||
        SQLITE_OK != sqlite3_open_v2(":memory:", (sqlite3 **)db, flags, NULL))
        return false;
    // clang-format on

    SQLITE_EXEC(*db, (const char *)___src_database_memory_sql, err_not_init);
    SQLITE_EXEC(*db, __pragmas, err_pragma);
    return true;

err_not_init:
    *db = NULL;
    return false;

err_pragma:
    sqlite3_close((sqlite3 *)*db);
    *db = NULL;
    return false;
}

PRIVATE_FUNCTION bool
___attach(lkt_db *db, const char *name, const char *path)
{
    static const char SQL_ATTACH_TEMPLATE[] = "ATTACH '%s' AS '%s';";
    char SQL_ATTACH[LKT_MAX_SQLITE_STATEMENT];
    bool ret = false;

    safe_snprintf(SQL_ATTACH, LKT_MAX_SQLITE_STATEMENT, SQL_ATTACH_TEMPLATE, path, name);
    SQLITE_EXEC(db, SQL_ATTACH, err_no_attach);
    LOG_INFO("DB", "Attached database '%s' with path '%s'", name, path);
    ret = true;
err_no_attach:
    return ret;
}

PRIVATE_FUNCTION bool
___detach(lkt_db *db, const char *name)
{
    static const char SQL_DETACH_TEMPLATE[] = "DETACH '%s';\n";
    char SQL_DETACH[LKT_MAX_SQLITE_STATEMENT];
    bool ret = false;

    safe_snprintf(SQL_DETACH, LKT_MAX_SQLITE_STATEMENT, SQL_DETACH_TEMPLATE, name);
    SQLITE_EXEC(db, SQL_DETACH, err_no_detach);
    LOG_INFO("DB", "Detached database '%s'", name);
    ret = true;
err_no_detach:
    return ret;
}

PRIVATE_FUNCTION bool
__check_schema(lkt_db *db)
{
    static const char *SQL = "SELECT COUNT(name)"
                             " FROM sqlite_master"
                             " WHERE type='table' AND name IN"
                             " ('kara', 'playlist', 'kara_playlist', 'queue', 'users', 'stickers',"
                             "  '''stickers.kara''', '''stickers.plt''', 'updates', 'misc', "
                             "  'kara_type', 'kara_category', 'language');";
    sqlite3_stmt *stmt = NULL;
    SQLITE_PREPARE(db, stmt, SQL, error);
    SQLITE_STEP_ROW(db, stmt, error);
    if (9 != sqlite3_column_int(stmt, 0)) {
        sqlite3_finalize(stmt);
        return true;
    }
error:
    LOG_ERROR("DB", "Schema is invalid: missing tables");
    if (stmt)
        sqlite3_finalize(stmt);
    return false;
}

PRIVATE_FUNCTION bool
___is_attached(lkt_db *db, const char *name)
{
    static const char *SQL_STMT = "SELECT name FROM pragma_database_list WHERE name = ?;\n";
    sqlite3_stmt *stmt          = NULL;
    bool ret                    = false;

    SQLITE_PREPARE(db, stmt, SQL_STMT, error);
    SQLITE_BIND_TEXT(db, stmt, 1, name, error);
    SQLITE_STEP_ROW(db, stmt, error);
    ret = true;
error:
    sqlite3_finalize(stmt);
    return ret;
}

PRIVATE_FUNCTION void
___flush_cache_from_disk(lkt_db *db)
{
    static const char *SQL =
        "INSERT INTO kara_cache (kara_id, mtime, duration, file_path, magic_error)"
        " SELECT id, cached_mtime, cached_duration, file_path, 2"
        " FROM kara;";
    SQLITE_EXEC(db, SQL, error);
    LOG_INFO("DB", "Kara cache was flushed from disk");
    return;
error:
    LOG_ERROR("DB", "Failed to flush kara cache from disk");
    return;
}

bool
database_open(lkt_db *db, const char *dbpath, bool check)
{
    bool retry = false;
    if (___is_sql_str_invalid(dbpath)) {
        LOG_ERROR("DB", "The database path '%s' is invalid", dbpath);
        return false;
    }

    /* Try the simple attach, for already initialized database */
retry:
    if (!___attach(db, LKT_PROTECTED_DATABASE, dbpath))
        goto init;
    if (!__check_schema(db))
        goto init;
    ___inc(db, LKT_PROTECTED_DATABASE, check);

    LKT_DATABASE_VERSION version = database_get_version(db);
    switch (version) {
    case LKT_DATABASE_VERSION_ALPHA:
        LOG_ERROR(
            "DB",
            "The database is in 'alpha' version, lektor might break, consider upgrading your db!");
        break;

    case LKT_DATABASE_VERSION_MK_7_1:
        LOG_INFO("DB", "Opening database, version is 'mk-7.1'");
        ___flush_cache_from_disk(db);
        break;
    }
    return true;

    /* Need init */
init:
    if (retry) {
        LOG_ERROR("DB", "Retry to init database '%s', this is an error", dbpath);
        return false;
    }

    retry = true;
    LOG_WARN("DB", "Database '%s' is not correctly initialized, init it", dbpath);
    ___detach(db, LKT_PROTECTED_DATABASE);

    if (database_init(dbpath))
        goto retry;

    else {
        LOG_ERROR("DB", "Failed to init database '%s'", dbpath);
        return false;
    }
}

bool
database_attach(lkt_db *db, const char *name, const char *dbpath)
{
    RETURN_UNLESS(strcasecmp(LKT_PROTECTED_DATABASE, name),
                  " The database " LKT_PROTECTED_DATABASE " is protected, can't attach a "
                  "database with the same name",
                  false);

    if (___is_sql_str_invalid(name)) {
        LOG_ERROR("DB", "The database name '%s' is invalid", name);
        return false;
    }

    if (___is_attached(db, name)) {
        LOG_ERROR("DB", "The database '%s' is already attached", name);
        return false;
    }

    if (!___attach(db, name, dbpath)) {
        LOG_ERROR("DB", "Failed to attach database named '%s' with path '%s'", name, dbpath);
        return false;
    }

    LOG_INFO("DB", "Database '%s' attached", name);
    return true;
}

bool
database_detach(lkt_db *db, const char *name)
{
    RETURN_UNLESS(strcasecmp(LKT_PROTECTED_DATABASE, name),
                  " The database " LKT_PROTECTED_DATABASE " is protected, can't detach it", false);

    if (___is_sql_str_invalid(name)) {
        LOG_ERROR("DB", "The database name '%s' is invalid", name);
        return false;
    }

    if (!___is_attached(db, name)) {
        LOG_ERROR("DB", "The database '%s' is not attached", name);
        return false;
    }

    if (!___detach(db, name)) {
        LOG_ERROR("DB", "Failed to detach database named %s", name);
        return false;
    }

    LOG_INFO("DB", "Database '%s' detached", name);
    return true;
}

bool
database_init(const char *dbpath)
{
    sqlite3 *db;
    if (SQLITE_OK != sqlite3_open(dbpath, &db))
        goto error;
    SQLITE_EXEC(db, (const char *)___src_database_disk_sql, error);
    LOG_INFO("DB", "Initialized the 'disk' database successfully, path was '%s'", dbpath);
    return true;
error:
    LOG_ERROR("DB", "Failed to init the 'disk' database, path was '%s'", dbpath);
    return false;
}
