#define _POSIX_C_SOURCE 200808L

#include <lektor/lktconfig.h>
#include <lektor/common.h>
#include <lektor/config.h>
#include <lektor/net.h>
#include <lektor/cmd.h>
#include <lektor/reg.h>
#include <lektor/database.h>
#include <lektor/commands.h>
#include <lektor/segv.h>
#include <lektor/launch.h>

#include <wait.h>
#include <spawn.h>
#include <libgen.h>
#include <signal.h>
#include <termios.h>

#if defined(LKT_STATIC_MODULE)
REG_DECLARE(sdl2_reg)
REG_DECLARE(repo_reg)
#endif

/* Disable echo in the console, only for the program */
struct termios ___tty_cur, ___tty_save;
static int ___tty_must_be_restored = 0;

CONSTRUCTOR_FUNCTION
___tty_disable_echo(void)
{
    GOTO_IF((tcgetattr(STDIN_FILENO, &___tty_cur) == -1), "Failed to get termios flags", error);

    ___tty_save = ___tty_cur;
    ___tty_cur.c_lflag &= ~ECHO;

    GOTO_IF((tcsetattr(STDIN_FILENO, TCSAFLUSH, &___tty_cur) == -1),
            "Failed to turned of echo in termios flags", error);
    ___tty_must_be_restored = 1;

    return;
error:
    /* Fails silently for the moment */
    return;
}

DESTRUCTOR_FUNCTION
___tty_enable_echo(void)
{
    RETURN_UNLESS(___tty_must_be_restored, "No need to restore the tty", NOTHING);
    RETURN_IF(tcsetattr(STDIN_FILENO, TCSANOW, &___tty_save) == -1, "Failed to reset termios flags",
              NOTHING);
    errno = 0;
    RETURN_IF(fflush(stdin) != 0, "Failed to flush stdin", NOTHING);
}

/* Recursive mkdir, where the last word of the string is a file, not a folder. */
PRIVATE_FUNCTION void
__mkdir(const char *dir)
{
    char tmp[PATH_MAX];
    char *p = NULL;
    safe_snprintf(tmp, sizeof(tmp), "%s", dir);
    size_t len = strlen(tmp);
    /* In our case, the final word is always a file, not a folder. */
    if (tmp[len - 1] == '/')
        tmp[len - 1] = 0;
    for (p = tmp + 1; *p; p++) {
        if (*p == '/') {
            *p = 0;
            mkdir(tmp, 00700);
            *p = '/';
        }
    }
    /* Don't do final mkdir here, because in our case the final word in the string
     * is a file, not a folder.
     * mkdir(tmp, S_IRWXU); */
}

int
main(int argc, char *argv[])
{
    lkt_install_segv_handler();
    lkt_thread_set_name("lektord/daemon");

    REG_BEGIN(server_reg)
    REG_ADD(launch_ext_klkt)
    REG_ADD(launch_caching)
#if defined(LKT_STATIC_MODULE)
    REG_REGISTER("repo", repo_reg)
    REG_REGISTER("sdl2", sdl2_reg)
#endif
    REG_END()

    /* Enforce some locals */
    RETURN_UNLESS(setlocale(LC_ALL, "en_US.UTF-8"), "Failed to set LC_ALL to UTF-8", 1);
    RETURN_UNLESS(setlocale(LC_CTYPE, ""), "Failed to set LC_CTYPE", 1);
    RETURN_UNLESS(setlocale(LC_NUMERIC, "C"), "Failed to set LC_NUMERIC for mpv", 1);

    int autoclear;
    int opt;
    int check_exclusive = 1;
    int dump_and_abort  = 0;
    char *conf_file     = LKT_ALLOC_ARRAY(char, PATH_MAX);
    cmd_set_executable_name("lektord");

    /* Check args */
    while ((opt = getopt(argc, argv, "DhFf:")) != -1) {
        switch (opt) {
        case 'F': check_exclusive = 0; break;
        case 'f': strncpy(conf_file, optarg, PATH_MAX); break;
        case 'D': dump_and_abort = 1; break;
        case 'h':
        default: print_help(); exit(EXIT_SUCCESS);
        }
    }

    reg_export(server_reg);
    if (lkt_read_self_exe(exe, PATH_MAX))
        LOG_WARN("INIT", "Failed to read self executable path, restart may not work");
    cmd_set_executable_name(exe);
    LOG_WARN("INIT", "The argv[0] from main is: %s", exe);

    /* Resolves */
    resolve_appimage();
    resolve_path();

    /* Init the server */
    struct lkt_state srv;
    memset(&srv, 0, sizeof(struct lkt_state));
    srv.kara_prefix = kara_dir;

    FAIL_IF(lkt_queue_new(&srv.queue), "Failed to create server queue");
    FAIL_UNLESS(database_new(&srv.db), "Failed to init memory database");

    /* Read or create default config file. */
    int retry_config_once = 0;
retry_config:
    if (conf_file[0] == '\0' && config_detect_file(conf_file, PATH_MAX)) {
        FAIL_IF(retry_config_once, "Failed to find a config file");

        LOG_INFO("INIT", "Creating default config file");
        config_default_file(conf_file, PATH_MAX);
        __mkdir(conf_file); /* Create the folder for the file. */

        FILE *file_desc = fopen(conf_file, "w+");
        FAIL_UNLESS(file_desc, "Failed to open default config file '%s' to initialize it",
                    conf_file);
        config_default(file_desc);
        fclose(file_desc);

        LOG_INFO("INIT", "Default configuration file has been writen to %s", conf_file);
        retry_config_once = 1;
        goto retry_config;
    }

    FAIL_IF(config_new(srv.db, conf_file), "Failed to read the config");

    /* Dump and abort here, if we are dumping informations about the server */
    if (dump_and_abort) {
        fprintf(stderr, "Config file is: %s\n", conf_file);
        if (database_config_dump(srv.db, stderr))
            exit(EXIT_FAILURE);
        exit(EXIT_SUCCESS);
    }

    /* No longer need the config file */
    free(conf_file);

    /* Read the configuration. We already know that the config is valid */
    database_config_get(srv.db, "player", "autoclear", &autoclear);
    database_config_get(srv.db, "database", "kara_dir", kara_dir, PATH_MAX);
    database_config_get(srv.db, "server", "host", srv.host, HOST_NAME_MAX);
    database_config_get(srv.db, "server", "port", srv.port, 5);

    /* Quick check with an explicit error message, to remide the users to
     * create the kara folder */
    FAIL_IF(access(kara_dir, R_OK | W_OK), "No access in read / write for folder %s", kara_dir);

    /* Finish to initialize. */
    database_config_get(srv.db, "database", "db_path", db_path, PATH_MAX);
    FAIL_UNLESS(database_open(srv.db, db_path, check_exclusive), "Can't open database");

    if (kara_dir[strlen(kara_dir) - 1] != '/')
        strncat(kara_dir, "/", PATH_MAX - 1);
    srv.kara_prefix = kara_dir;

    database_config_queue_default(srv.db);
    if (!env_get(LKT_ENV_RESTART) && autoclear) {
        LOG_INFO("INIT", "Clear queue, autoclear is set in ini file");
        database_queue_clear(srv.db);
    }
    lkt_queue_make_available(&srv.queue, LKT_EVENT_PROP);
    lkt_queue_make_available(&srv.queue, LKT_EVENT_UPDATE);

    /* Module initialization */
    {
        char module[PATH_MAX];

        database_config_get(srv.db, "player", "module", module, PATH_MAX);
        reg_import(module, &srv.window_mod.reg, &srv.window_mod.handle);

        database_config_get(srv.db, "repo", "module", module, PATH_MAX);
        reg_import(module, &srv.repo_mod.reg, &srv.repo_mod.handle);

        FAIL_IF(MOD_CALL(srv.repo_mod, "new", &srv.queue, srv.db), "Can't init repo module");
        FAIL_IF(MOD_CALL(srv.window_mod, "new", &srv.queue, srv.db), "Can't init player module");
    }

    /* Get ENV */
    {
        char *env_current = env_get(LKT_ENV_CURRENT);
        if (env_current && env_current[0] && !STR_MATCH(env_current, "NULL")) {
            LOG_INFO("INIT", "Restart playback from %s", env_current);
            lkt_queue_send(&srv.queue, LKT_EVENT_PLAY_POS,
                           (void *)(size_t)strtol(env_current, NULL, 0));
        }
    }

    LOG_INFO("INIT", "Lektor was %s", env_get(LKT_ENV_RESTART) ? "restarted" : "started");
    config_handle_hook(srv.db, "launched");
    lkt_listen(&srv);
    LOG_FATAL("The lkt_listen function returned, it shouldn't happen");
    return EXIT_FAILURE;
}
