#define _POSIX_C_SOURCE 200809L

#include <lektor/bufferfd.h>
#include <lektor/mkv.h>
#include <lektor/common.h>
#include <math.h>

#define MKV_TAG_MAX 64

#define EBML_ROOT             0x1a45dfa3
#define EBML_MKV_SEGMENT      0x18538067
#define EBML_MKV_SEEKHEAD     0x114d9b74
#define EBML_MKV_SEEK         0x00004dbb
#define EBML_MKV_SEEK_ID      0x000053ab
#define EBML_MKV_SEEK_POS     0x000053ac
#define EBML_MKV_TAGS         0x1254c367
#define EBML_MKV_TAG          0x00007373
#define EBML_MKV_TAG_TARGETS  0x000063c0
#define EBML_MKV_TAG_TTV      0x000068ca /* Tag: TargetTypeValue, L4 */
#define EBML_MKV_TAG_SIMPLE   0x000067c8 /* Tag: SimpleTag, L3+      */
#define EBML_MKV_TAG_NAME     0x000045a3
#define EBML_MKV_TAG_STRING   0x00004487
#define EBML_MKV_TAG_BINARY   0x00004485
#define EBML_MKV_INFO         0x1549a966 /* Segment: Info             L1 */
#define EBML_MKV_SEG_DURATION 0x00004489 /* Segment: Duration         L2 */
#define EBML_MKV_SEG_TS_SCALE 0x002ad7b1 /* Segment: TimestampScale   L2 */
#define EBML_MKV_SEG_NAME     0x00007ba9 /* Segment: Title            L2 */
#define EBML_MKV_ATTACHEMENT  0x1941a469
#define EBML_MKV_ATTCH_NAME   0x0000466e /* Attachement: FileName, L3        */
#define EBML_MKV_ATTCH_DESC   0x0000467e /* Attachement: FileDescription, L3 */
#define EBML_MKV_ATTCH_MIME   0x00004660 /* Attachement: FileMimeType, L3    */

#define EBML_MKV_CRC32 0x0000000bf /* CRC-32, they skip it in mpv (demux/ebml.c:463 aprox) */
#define EBML_MKV_VOID  0x0000000ec /* VOID element */

/* mkv_read_* functions take data from the bufferfd and parse the next "*" (one
 * of the following):
 *
 * - mkv_read_element_id reads the id of an element (up to 4 bytes).
 * - mkv_read_element_length reads the size of an element (up to 8 bytes).
 * - mkv_read_<TYPE> reads the size of the <TYPE> element and then the data
 *   according to the spec (cf. header file).
 *
 * mkv_read_utf8 differs from mkv_read_binary, the former inserts a null byte
 * after the data (so it effectively reads one less byte than the later).
 *
 * mkv_skip_element reads the element size and then skips this amount of bytes.
 */

static ssize_t
mkv_read_element_id(struct bufferfd *bf, uint32_t *res)
{
    uint8_t data[4];
    uint8_t nbytes;

    if (bufferfd_byte(bf, data) < 0)
        return -1;

    for (nbytes = 0; nbytes < 4; nbytes++) {
        if (data[0] & (0x80u >> nbytes))
            break;
    }

    if (nbytes == 4 || bufferfd_bytes(bf, nbytes, data + 1) < 0)
        return -1;

    *res = 0;

    for (uint8_t i = 0; i <= nbytes; i++)
        *res = (*res << 8u) | data[i];

    return nbytes + 1;
}

static ssize_t
mkv_read_element_length(struct bufferfd *bf, uint64_t *res)
{
    uint8_t data[8];
    uint8_t nbytes;

    if (bufferfd_byte(bf, data) < 0)
        return -1;

    for (nbytes = 0; nbytes < 8; nbytes++) {
        uint8_t mask = 0x80u >> nbytes;

        if (data[0] & mask) {
            data[0] &= (uint8_t)~mask;
            break;
        }
    }

    if (nbytes == 8 || bufferfd_bytes(bf, nbytes, data + 1) < 0)
        return -1;

    *res = 0;

    for (uint8_t i = 0; i <= nbytes; i++)
        *res = (*res << 8u) | data[i];

    return nbytes + 1;
}

static ssize_t
mkv_read_uint(struct bufferfd *bf, uint64_t *res)
{
    uint8_t data[8];
    uint64_t data_len;
    ssize_t n;

    if ((n = mkv_read_element_length(bf, &data_len)) < 0 || 8 < data_len)
        return -1;

    if (data_len == 0)
        *res = 0;
    else {
        if (bufferfd_bytes(bf, data_len, data) < 0)
            return -1;

        *res = be_uint64_t(data, data_len);
    }

    return n + (ssize_t)data_len;
}

/* From https://www.matroska.org/technical/specs/index.html:
   Float - Big-endian, defined for 4 and 8 octets (32, 64 bits).
   Here we always return a double.
   Keep the LOG_DEBUG around if one day there is a need to debug that
   thing again... */
static ssize_t
mkv_read_float(struct bufferfd *bf, double *res)
{
    uint8_t data[8];
    uint64_t data_len;
    ssize_t n;

    if ((n = mkv_read_element_length(bf, &data_len)) < 0 || 8 < data_len)
        return -1;

    if (data_len == 0)
        *res = 0.;
    else {
        if (bufferfd_bytes(bf, data_len, data) < 0)
            return -1;

        if (data_len == 8) {
            *res = be_double_t(data);
#if 0
            LOG_DEBUG("MKV", "Float is of length %zu, brut data is %x%x%x%x%x%x%x%x, interpreted data is %f%s",
                      data_len, data[7], data[6], data[5], data[4], data[3], data[2], data[1], data[0], *res,
                      isnan(*res) ? ", is NAN!" : "");
#endif
        }

        else if (data_len == 4) {
            *res = (double)be_float_t(data);
#if 0
            LOG_DEBUG("MKV", "Float is of length %zu, brut data is %x%x%x%x, interpreted data is %f%s", data_len,
                      data[3], data[2], data[1], data[0], *res, isnan(*res) ? ", is NAN!" : "");
#endif
        }

        else
            return -1;
    }

    return n + (ssize_t)data_len;
}

/* Kept as a reference
static ssize_t
mkv_read_binary(struct bufferfd * bf, size_t max, uint8_t * res)
{
  uint64_t data_len;
  ssize_t n;
  size_t to_read;

  if ((n = mkv_read_element_length(bf, &data_len)) < 0)
    return -1;

  to_read = max < data_len ? max : (size_t) data_len;

  if (bufferfd_bytes(bf, to_read, res) < 0)
    return -1;

  bufferfd_skip(bf, data_len - to_read);
  return n + (ssize_t) data_len;
}
*/

static ssize_t
mkv_read_utf8(struct bufferfd *bf, size_t max, char *res)
{
    uint64_t data_len;
    ssize_t n;
    size_t to_read;

    if ((n = mkv_read_element_length(bf, &data_len)) < 0)
        return -1;

    max--;
    to_read = max < data_len ? max : (size_t)data_len;

    if (bufferfd_bytes(bf, to_read, (uint8_t *)res) < 0)
        return -1;

    res[to_read] = 0;
    bufferfd_skip(bf, data_len - to_read);
    return n + (ssize_t)data_len;
}

static ssize_t
mkv_skip_element(struct bufferfd *bf)
{
    uint64_t data_len;
    ssize_t n;

    if ((n = mkv_read_element_length(bf, &data_len)) < 0)
        return -1;

    bufferfd_skip(bf, data_len);
    return n + (ssize_t)data_len;
}

static ssize_t
mkv_skip_header(struct bufferfd *bf)
{
    uint32_t eid;
    uint64_t elength;
    ssize_t n;
    ssize_t consumed;

    if ((consumed = mkv_read_element_id(bf, &eid)) < 0 || eid != EBML_ROOT)
        return -1;

    if ((n = mkv_skip_element(bf)) < 0)
        return -1;

    consumed += n;

    if ((n = mkv_read_element_id(bf, &eid)) < 0 || eid != EBML_MKV_SEGMENT)
        return -1;

    consumed += n;

    if ((n = mkv_read_element_length(bf, &elength)) < 0)
        return -1;

    consumed += n;
    return consumed;
}

struct mkv_seek {
    uint32_t id;
    uint64_t pos;
};

static ssize_t
mkv_read_seek(struct bufferfd *bf, struct mkv_seek *res)
{
    ssize_t consumed;
    uint64_t data_len;
    int got_id  = 0;
    int got_pos = 0;

    if ((consumed = mkv_read_element_length(bf, &data_len)) < 0)
        return -1;

    consumed += (ssize_t)data_len;

    while (data_len > 0) {
        ssize_t n;
        uint32_t eid;
        uint64_t elength;

        if ((n = mkv_read_element_id(bf, &eid)) < 0)
            return -1;

        data_len -= n;

        if (eid == EBML_MKV_SEEK_ID) {
            uint8_t seek_id[4];

            if ((n = mkv_read_element_length(bf, &elength)) < 0 || 4 < elength)
                return -1;

            if (bufferfd_bytes(bf, elength, seek_id) < 0)
                return -1;

            res->id = be_uint32_t(seek_id, elength);
            data_len -= n + elength;
            got_id = 1;
        } else if (eid == EBML_MKV_SEEK_POS) {
            if ((n = mkv_read_uint(bf, &res->pos)) < 0)
                return -1;

            data_len -= n;
            got_pos = 1;
        } else {
            if ((n = mkv_skip_element(bf)) < 0)
                return -1;

            data_len -= n;
        }
    }

    if (!got_id || !got_pos)
        return -1;

    return consumed;
}

struct mkv_simple_tag {
    char name[LEKTOR_TAG_MAX];
    char value[LEKTOR_TAG_MAX];
};

static ssize_t
mkv_read_simple_tag(struct bufferfd *bf, struct mkv_simple_tag *res)
{
    ssize_t consumed;
    uint64_t data_len;
    uint32_t eid;

    if ((consumed = mkv_read_element_length(bf, &data_len)) < 0)
        return -1;

    consumed += (ssize_t)data_len;
    res->name[0]  = 0;
    res->value[0] = 0;

    while (data_len > 0) {
        ssize_t n;
        n = mkv_read_element_id(bf, &eid);

        if (n < 0)
            return -1;

        data_len -= n;

        if (eid == EBML_MKV_TAG_NAME)
            n = mkv_read_utf8(bf, LEKTOR_TAG_MAX, res->name);
        else if (eid == EBML_MKV_TAG_STRING || eid == EBML_MKV_TAG_BINARY)
            n = mkv_read_utf8(bf, LEKTOR_TAG_MAX, res->value);
        else
            n = mkv_skip_element(bf);

        if (n < 0)
            return -1;

        data_len -= (size_t)n < data_len ? (size_t)n : data_len;
    }

    return consumed;
}

static ssize_t
mkv_read_targets(struct bufferfd *bf, uint64_t *target_type_value)
{
    ssize_t consumed;
    uint64_t data_len;

    if ((consumed = mkv_read_element_length(bf, &data_len)) < 0)
        return -1;

    consumed += (ssize_t)data_len;

    while (data_len > 0) {
        ssize_t n;
        uint32_t eid;

        if ((n = mkv_read_element_id(bf, &eid)) < 0)
            return -1;

        data_len -= n;

        if (eid == EBML_MKV_TAG_TTV) {
            if ((n = mkv_read_uint(bf, target_type_value)) < 0)
                return -1;

            data_len -= n;
            break;
        }

        if ((n = mkv_skip_element(bf)) < 0)
            return -1;

        data_len -= (size_t)n < data_len ? (size_t)n : data_len;
    }

    bufferfd_skip(bf, data_len);
    return consumed;
}

struct mkv_tag {
    uint64_t target_type_value;
    struct mkv_simple_tag simples[MKV_TAG_MAX];
};

static ssize_t
mkv_read_tag(struct bufferfd *bf, struct mkv_tag *res)
{
    ssize_t consumed;
    ssize_t n;
    uint64_t data_len;
    res->target_type_value = 50;
    memset(res->simples, 0, MKV_TAG_MAX * sizeof(struct mkv_simple_tag));

    if ((consumed = mkv_read_element_length(bf, &data_len)) < 0)
        return -1;

    consumed += data_len;

    for (size_t i = 0; i < MKV_TAG_MAX && data_len > 0;) {
        uint32_t eid;

        if ((n = mkv_read_element_id(bf, &eid)) < 0)
            return -1;

        data_len -= n;

        if (eid == EBML_MKV_TAG_TARGETS) {
            if ((n = mkv_read_targets(bf, &res->target_type_value)) < 0)
                return -1;
        } else if (eid == EBML_MKV_TAG_SIMPLE) {
            if ((n = mkv_read_simple_tag(bf, &res->simples[i])) < 0)
                return -1;

            i++;
        } else {
            if ((n = mkv_skip_element(bf)) < 0)
                return -1;
        }

        data_len -= n;
    }

    bufferfd_skip(bf, data_len);
    return consumed;
}

static ssize_t
mkv_seek(struct bufferfd *bf, size_t header_len, size_t what)
{
    uint64_t seekhead_len;

    if (mkv_read_element_length(bf, &seekhead_len) < 0)
        return -1;

    while (seekhead_len > 0) {
        struct mkv_seek s;
        s.id  = 0;
        s.pos = 0;
        uint32_t eid;
        ssize_t n;

        if ((n = mkv_read_element_id(bf, &eid)) < 0)
            return -1;

        seekhead_len -= n;

        if (eid == EBML_MKV_SEEK)
            n = mkv_read_seek(bf, &s);
        else
            n = mkv_skip_element(bf);

        if (n < 0)
            return -1;

        seekhead_len -= n;

        if (eid == EBML_MKV_SEEK && s.id == what)
            return bufferfd_seek(bf, s.pos + header_len);
    }

    return 0;
}

static void
kara_add_simple_tag(struct kara_metadata *dst, struct mkv_simple_tag *t, uint64_t target_type_value)
{
    if (STR_MATCH("TITLE", t->name)) {
        if (target_type_value == 50)
            strncpy(dst->song_name, t->value, LEKTOR_TAG_MAX - 1);

        else if (target_type_value == 70)
            strncpy(dst->source_name, t->value, LEKTOR_TAG_MAX - 1);
    }

    else if (STR_MATCH("CONTENT_TYPE", t->name))
        strncpy(dst->category, t->value, LEKTOR_TAG_MAX - 1);

    else if (STR_MATCH("ADDRESS", t->name))
        strncpy(dst->language, t->value, LEKTOR_TAG_MAX - 1);

    else if (STR_MATCH("ARTIST", t->name))
        strncpy(dst->author_name, t->value, LEKTOR_TAG_MAX - 1);

    else if (STR_MATCH("GENRE", t->name))
        strncpy(dst->song_type, t->value, LEKTOR_TAG_MAX - 1);

    else if (STR_MATCH("TRACK", t->name))
        dst->song_number = (int)strtol(t->value, NULL, 10);
}

static void
kara_add_tag(struct kara_metadata *dst, struct mkv_tag *t)
{
    for (size_t i = 0; i < MKV_TAG_MAX; i++) {
        if (t->simples[i].name[0] == 0)
            break;

        kara_add_simple_tag(dst, &t->simples[i], t->target_type_value);
    }
}

static int
kara_read_tags(struct bufferfd *bf, struct kara_metadata *dst)
{
    uint64_t data_len;

    if (mkv_read_element_length(bf, &data_len) < 0)
        return -1;

    while (data_len > 0) {
        ssize_t n;
        uint32_t eid;
        struct mkv_tag t;

        if ((n = mkv_read_element_id(bf, &eid)) < 0 || eid != EBML_MKV_TAG)
            return -1;

        data_len -= n;

        if ((n = mkv_read_tag(bf, &t)) < 0)
            return -1;

        kara_add_tag(dst, &t);
        data_len -= n;
    }

    return 0;
}

static int
kara_read_segment_info(struct bufferfd *bf, double *len)
{
    uint64_t data_len;

    if (mkv_read_element_length(bf, &data_len) < 0)
        return -1;

    while (data_len > 0) {
        ssize_t n;
        uint32_t eid;

        if ((n = mkv_read_element_id(bf, &eid)) < 0)
            return -1;

        data_len -= n;

        if (eid == EBML_MKV_SEG_DURATION) {
            if ((n = mkv_read_float(bf, len)) < 0)
                return -1;
            LOG_DEBUG("MKV", "Got segment duration %f", *len);
        }

#if 0
        /* Dividing the EBML_MKV_SEG_DURATION by 10 should give the time in seconds */
        uint64_t scale = 1;
        else if (eid == EBML_MKV_SEG_TS_SCALE) {
            if ((n = mkv_read_uint(bf, &scale)) < 0)
                return -1;
            LOG_DEBUG("MKV", "Got segment ts scale %zu", scale);
        }
#endif

        else if ((n = mkv_skip_element(bf)) < 0)
            return -1;

        data_len -= n;
    }

    *len /= 10; /* In s */
    return 0;
}

int
kara_read_length(double *len, const char *filename)
{
    errno                = 0;
    int status_code      = -1;
    struct bufferfd file = {
        .len = 0,
        .pos = 0,
        .fd  = open(filename, O_RDONLY),
    };

    if (file.fd < 0) {
        LOG_ERROR("MKV", "Faield to open file '%s': %s", filename, strerror(errno));
        return -1;
    }

    uint32_t eid;
    uint64_t elength;
    ssize_t header_len = mkv_skip_header(&file);

    if (header_len < 0)
        goto error;

    *len = 0.;

    for (;;) {
        if (mkv_read_element_id(&file, &eid) < 0)
            goto error;

        if (eid == EBML_MKV_SEEKHEAD) {
            if (mkv_seek(&file, (size_t)header_len, (size_t)EBML_MKV_SEGMENT) < 0)
                goto error;
        }

        else if (eid == EBML_MKV_INFO) {
            if (kara_read_segment_info(&file, len) < 0)
                goto error;
            break;
        }

        else {
            if (mkv_read_element_length(&file, &elength) < 0)
                goto error;
            if (bufferfd_skip(&file, elength) < 0)
                goto error;
        }
    }

    status_code = 0;
error:
    close(file.fd);
    if (fabs(*len) < 1)
        LOG_ERROR("MKV", "Duration for file %s is %f, might be an error!", filename, *len);
    return status_code;
}

int
kara_metadata_read(struct kara_metadata *dst, const char *filename)
{
    int status_code = -1;
    struct bufferfd file;
    errno    = 0;
    file.len = 0;
    file.pos = 0;
    file.fd  = open(filename, O_RDONLY);

    if (file.fd < 0) {
        LOG_ERROR("MKV", "Failed to open file '%s': %s", filename, strerror(errno));
        return -1;
    }

    uint32_t eid;
    uint64_t elength;
    ssize_t header_len = mkv_skip_header(&file);

    if (header_len < 0)
        goto error;

    memset(dst, 0, sizeof(struct kara_metadata));

    for (;;) {
        if (mkv_read_element_id(&file, &eid) < 0)
            goto error;

        if (eid == EBML_MKV_SEEKHEAD) {
            if (mkv_seek(&file, (size_t)header_len, (size_t)EBML_MKV_TAGS) < 0)
                goto error;
        } else if (eid == EBML_MKV_TAGS) {
            if (kara_read_tags(&file, dst) < 0)
                goto error;

            break;
        } else {
            if (mkv_read_element_length(&file, &elength) < 0)
                goto error;

            if (bufferfd_skip(&file, elength) < 0)
                goto error;
        }
    }

    status_code = 0;
error:
    close(file.fd);
    return status_code;
}
