#define _POSIX_C_SOURCE 200809L

#include "mpv.h"

#include <lektor/common.h>
#include <lektor/database.h>

void
lmpv_free(mpv_handle **ctx)
{
    RETURN_UNLESS(ctx && *ctx, "Missing mpv ctx", NOTHING);
    static const char *cmd[] = { "quit", NULL };
    mpv_command(*ctx, cmd);
    mpv_destroy(*ctx);
    *ctx = NULL;
    LOG_INFO("WINDOW", "The mpv context was destroyed");
}

mpv_handle *
lmpv_prepare(lkt_db *db)
{
    mpv_handle *ctx = mpv_create();
    int status;
    RETURN_UNLESS(ctx, "Failed to create context", NULL);
    char ___opt[LKT_LINE_MAX];

#define MPV_SET_OPTION_PASS_ERRORS(opt, value) mpv_set_option_string(ctx, opt, value)
#define MPV_SET_OPTION(opt, value)                                                               \
    if ((status = MPV_SET_OPTION_PASS_ERRORS(opt, value)) < 0) {                                 \
        LOG_ERROR("WINDOW", "Failed to set %s to %s: %s", opt, value, mpv_error_string(status)); \
        return NULL;                                                                             \
    }
#define MPV_SET_FROM_INI(opt, section, key)                                     \
    if (!database_config_get(db, section, key, ___opt, LKT_LINE_MAX)) {         \
        LOG_WARN("WINDOW", "Failed to get option " key " in section " section); \
        return ctx;                                                             \
    }                                                                           \
    MPV_SET_OPTION(opt, ___opt);
#define MPV_SET_OPTION_COND(opt, value, ini_section, ini_option, def)                \
    ({                                                                               \
        int ___cond;                                                                 \
        if (!database_config_get(db, ini_section, ini_option, &___cond)) {           \
            LOG_WARN("WINDOW", "Default [" ini_section "/" ini_option "] to " #def); \
            ___cond = (def);                                                         \
        }                                                                            \
        if (___cond)                                                                 \
            MPV_SET_OPTION_PASS_ERRORS(opt, value);                                  \
    })

    // clang-format off
    MPV_SET_OPTION("script-opts",               "autoload-disabled=yes");
    MPV_SET_OPTION("load-scripts",              "no");
    MPV_SET_OPTION("input-default-bindings",    "yes");
    MPV_SET_OPTION("input-vo-keyboard",         "yes");
    MPV_SET_OPTION("replaygain",                "track");
    MPV_SET_OPTION("demuxer-readahead-secs",    "5.0");
    MPV_SET_OPTION("demuxer-max-bytes",         "100M");
    MPV_SET_OPTION("hwdec",                     "yes");

    MPV_SET_OPTION_COND("gpu-context", "x11", "player", "force_x11", 1); /* Wayland you sucks */

    MPV_SET_FROM_INI("osd-font-size", "player", "font_size");
    MPV_SET_FROM_INI("osd-font",      "player", "font_name");
    MPV_SET_FROM_INI("osd-duration",  "player", "msg_duration");
    // clang-format on
    return ctx;
#undef MPV_SET_OPTION
#undef MPV_SET_FROM_INI
#undef MPV_SET_OPTION_COND
}

int
lmpv_observe_properties(mpv_handle *ctx)
{
    return (mpv_observe_property(ctx, 0, "ao-volume", MPV_FORMAT_INT64) >= 0) &&
           (mpv_observe_property(ctx, 0, "duration", MPV_FORMAT_INT64) >= 0) &&
           (mpv_observe_property(ctx, 0, "time-pos", MPV_FORMAT_INT64) >= 0) &&
           (mpv_observe_property(ctx, 0, "pause", MPV_FORMAT_FLAG) >= 0);
}

mpv_handle *
lmpv_new(unsigned long int wid, lkt_db *db)
{
    mpv_handle *ctx = lmpv_prepare(db);
    int status;

    if ((status = mpv_set_property(ctx, "wid", MPV_FORMAT_INT64, &wid)) < 0) {
        LOG_ERROR("WINDOW", "Failed to set wid: %s", mpv_error_string(status));
        goto error;
    }

    if ((status = mpv_initialize(ctx)) < 0) {
        LOG_ERROR("WINDOW", "Failed to init mpv: %s", mpv_error_string(status));
        goto error;
    }

    if (!lmpv_observe_properties(ctx)) {
        LOG_ERROR("WINDOW", "Failed to observe properties");
        goto error;
    }

    LOG_INFO("WINDOW", "Create mpv context");
    return ctx;
error:
    lmpv_free(&ctx);
    return NULL;
}

int
lmpv_set_position(mpv_handle *ctx, int pos)
{
    RETURN_UNLESS(ctx, "Missing mpv ctx", 1);
    int status;
    char str[LKT_LINE_MAX];
    memset(str, 0, LKT_LINE_MAX * sizeof(char));
    safe_snprintf(str, LKT_LINE_MAX, "%d", pos);
    const char *cmd[] = { "seek", str, "absolute", NULL };
    if ((status = mpv_command_async(ctx, 0, cmd)) < 0) {
        LOG_ERROR("WINDOW", "Failed to execute command: %s", mpv_error_string(status));
        return 1;
    }
    return 0;
}

int
lmpv_set_volume(mpv_handle *ctx, int vol)
{
    RETURN_UNLESS(ctx, "Missing mpv ctx", 1);
    int status;
    char str[5];
    memset(str, 0, 5);
    safe_snprintf(str, 4, "%d", vol);
    const char *cmd[] = { "set", "ao-volume", str, NULL };
    if ((status = mpv_command_async(ctx, 0, cmd)) < 0) {
        LOG_ERROR("WINDOW", "Failed to execute command: %s", mpv_error_string(status));
        return 1;
    }
    return 0;
}

int
lmpv_load_file(mpv_handle *ctx, const char *file)
{
    RETURN_UNLESS(ctx, "Missing mpv ctx", 1);
    RETURN_IF(access(file, R_OK), "Failed to read file", 1);
    const char *cmd1[] = { "loadfile", file, "replace", NULL };
    const char *cmd2[] = { "set", "pause", "0", NULL };
    int status;
    if ((status = mpv_command_async(ctx, 0, cmd1)) < 0) {
        LOG_ERROR("WINDOW", "Failed to add '%s': %s", file, mpv_error_string(status));
        return 1;
    }
    if ((status = mpv_command_async(ctx, 0, cmd2)) < 0) {
        LOG_ERROR("WINDOW", "Failed to set state to play: %s", mpv_error_string(status));
        return 1;
    }
    return 0;
}

int
lmpv_toggle_pause(mpv_handle *ctx)
{
    RETURN_UNLESS(ctx, "Missing mpv ctx", 1);
    const char *cmd[] = { "cycle", "pause", "up", NULL };
    int status;
    if ((status = mpv_command_async(ctx, 0, cmd)) < 0) {
        LOG_ERROR("WINDOW", "Failed issus command: %s", mpv_error_string(status));
        return 1;
    }
    return 0;
}

int
lmpv_handle(struct lkt_module *mod, mpv_handle *ctx, struct queue *queue,
            struct lmpv_handle_arg *arg)
{
    size_t ao_volume;
    mpv_event *event = NULL;
    mpv_event_property *prop;
    RETURN_UNLESS(ctx, "Invalid argument", 1);

loop:
    event = mpv_wait_event(ctx, 0);

    switch (event->event_id) {
    case MPV_EVENT_PAUSE:
        SET_STATE_FLAG(*(arg->state), PAUSE);
        lkt_queue_send(queue, LKT_EVENT_PLAY_TOGGLE, LKT_PLAY_PAUSE);
        break;

    case MPV_EVENT_UNPAUSE:
        SET_STATE_FLAG(*(arg->state), PLAY);
        lkt_queue_send(queue, LKT_EVENT_PLAY_TOGGLE, LKT_PLAY_PLAY);
        break;

    case MPV_EVENT_SHUTDOWN:
        *(arg->time_pos)      = 0;
        *(arg->time_duration) = 0;
        SET_STATE_FLAG(*(arg->state), STOP);
        lkt_queue_send(queue, LKT_EVENT_PLAY_TOGGLE, LKT_PLAY_STOP);
        reg_call(mod->reg, "close", 1, mod->data);
        return 1;

    case MPV_EVENT_NONE: return 1;

    case MPV_EVENT_START_FILE:
        SET_STATE_FLAG(*(arg->state), PLAY);
        *(arg->hinib) = false;
        LOG_DEBUG("WINDOW", "Start of file!");
        break;

    case MPV_EVENT_END_FILE:
        LOG_DEBUG("WINDOW", "End of file!");
        if (((*(arg->state)) != STATE_STOP) && !(*(arg->hinib))) {
            LOG_DEBUG("WINDOW", "Send play_next event");
            lkt_queue_send(queue, LKT_EVENT_PLAY_NEXT, NULL);
        } else
            LOG_DEBUG("WINDOW", "Don't send play_next event, hinib is %d and state is %s",
                      *(arg->hinib),
                      (*(arg->state) == STATE_STOP)    ? "STOP"
                      : (*(arg->state) == STATE_PLAY)  ? "PLAY"
                      : (*(arg->state) == STATE_PAUSE) ? "PAUSE"
                                                       : "UNKNOWN");
        break;

    case MPV_EVENT_PROPERTY_CHANGE:
        prop = (mpv_event_property *)event->data;
        if (prop->format == MPV_FORMAT_NONE)
            break;
        /* MPV volume (BUG: The flag is not MPV_FORMAT_NONE only at the end of the song...) */
        if (STR_MATCH(prop->name, "ao-volume") && prop->format == MPV_FORMAT_INT64) {
            ao_volume = *(int *)prop->data;
            lkt_queue_send(queue, LKT_EVENT_PROP_VOL, (void *)ao_volume);
        }
        /* File duration */
        if (STR_MATCH(prop->name, "duration") && prop->format == MPV_FORMAT_INT64) {
            *(arg->time_duration) = *(int *)prop->data;
            lkt_queue_send(queue, LKT_EVENT_PROP_DUR, (void *)(size_t) * (arg->time_duration));
        }
        if (STR_MATCH(prop->name, "time-pos") && prop->format == MPV_FORMAT_INT64) {
            *(arg->time_pos) = *(int *)prop->data;
            lkt_queue_send(queue, LKT_EVENT_PROP_TIME, (void *)(size_t) * (arg->time_pos));
        }
        /* Pause state */
        if (STR_MATCH(prop->name, "pause") && prop->format == MPV_FORMAT_FLAG) {
            if (*(bool *)prop->data) {
                lkt_queue_send(queue, LKT_EVENT_PLAY_TOGGLE, LKT_PLAY_PAUSE);
                SET_STATE_FLAG(*(arg->state), PAUSE);
            } else {
                lkt_queue_send(queue, LKT_EVENT_PLAY_TOGGLE, LKT_PLAY_PLAY);
                SET_STATE_FLAG(*(arg->state), PLAY);
            }
        }
        break;

    case MPV_EVENT_COMMAND_REPLY: free((void *)event->reply_userdata); break;

    /* Ignored */
    case MPV_EVENT_VIDEO_RECONFIG:
    case MPV_EVENT_AUDIO_RECONFIG:
    case MPV_EVENT_IDLE:
    case MPV_EVENT_TRACKS_CHANGED:
    case MPV_EVENT_FILE_LOADED:
    case MPV_EVENT_METADATA_UPDATE:
    case MPV_EVENT_PLAYBACK_RESTART: break;

    /* Forgot to ignore an event, or it should not be ignored and should be
     * handled, or a new event is present in the API */
    default: LOG_WARN("WINDOW", "Unhandled mpv event '%s'", mpv_event_name(event->event_id)); break;
    }

    /* Can I seek? */
    if (*(arg->set_seek) >= 0 && *(arg->state) != STATE_STOP) {
        if (lmpv_set_position(ctx, *(arg->set_seek)))
            LOG_DEBUG("WINDOW", "Failed to seek the mpv context to %s", *(arg->set_seek));
        else
            LOG_DEBUG("WINDOW", "Seeking the mpv context to %s", *(arg->set_seek));
        *(arg->set_seek) = -1;
    }

    /* A loop without indentation. Ugly but better for not-really-wide screens */
    goto loop;
}
