#define _POSIX_C_SOURCE 200809L

#include "worker.h"
#include <lektor/common.h>
#include <sys/sysinfo.h>
#include <sched.h>

static void *
___worker_thread(void *___pool)
{
    lkt_thread_set_name("lektord/worker");

    volatile struct worker_pool *pool = (volatile struct worker_pool *)___pool;
    volatile void *volatile arg;
    worker_function func;
    FOR_EVER
    {
        assert(!pthread_mutex_lock((pthread_mutex_t *)&pool->lock));
        if (pool->len) {
            --(pool->len);
            ++(pool->thread_working);
            func = pool->functions[pool->len];
            arg  = pool->args[pool->len];
        } else if (pool->exit) {
            assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
            LOG_INFO("WORKER", "Exiting");
            break;
        } else {
            assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
            sleep(1);
            continue;
        }

        assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
        LOG_INFO("WORKER", "Picked up a function");
        func(___pool, (void *)arg);
        LOG_INFO("WORKER", "Finished work for a function");

        assert(!pthread_mutex_lock((pthread_mutex_t *)&pool->lock));
        --(pool->thread_working);
        assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
    }
    pthread_exit(NULL);
}

WORKER_STATUS
worker_pool_get_status(void *__pool)
{
    volatile struct worker_pool *pool = (volatile struct worker_pool *)__pool;
    assert(!pthread_mutex_lock((pthread_mutex_t *)&pool->lock));
    WORKER_STATUS ret = WORKER_STATUS_NONE;

    /* Case handling to get the correct status, simple for the moment, but may
     * involve some logic in the future. */
    if (pool->exit && pool->len == 0)
        ret = WORKER_STATUS_INTERRUPT;

    assert(!pthread_mutex_unlock((pthread_mutex_t *)&pool->lock));
    return ret;
}

void
worker_pool_interrupt(struct worker_pool *pool)
{
    assert(!pthread_mutex_lock(&pool->lock));
    pool->len  = 0;
    pool->exit = 1;
    assert(!pthread_mutex_unlock(&pool->lock));
}

int
worker_pool_new(struct worker_pool *ret, size_t init_size, size_t thread_count)
{
    if (!thread_count)
        thread_count = get_nprocs_conf();
    struct worker_pool __ret = {
        .functions   = LKT_ALLOC_ARRAY(worker_function, init_size),
        .args        = LKT_ALLOC_ARRAY(volatile void *, init_size),
        .threads     = LKT_ALLOC_ARRAY(pthread_t, thread_count),
        .size        = init_size,
        .thread_size = thread_count,
        .len         = 0u,
        .exit        = 0,
    };
    *ret = __ret;
    assert(!pthread_mutex_init(&ret->lock, NULL));
    size_t i;
    for (i = 0; i < ret->thread_size; ++i)
        assert(!pthread_create(&ret->threads[i], NULL, ___worker_thread, ret));
    return 0;
}

void
worker_pool_free(struct worker_pool *pool)
{
    LOG_DEBUG("WORKER", "Freeing worker pool %p", pool);
    size_t i;
    pool->exit = 1;
    for (i = 0; i < pool->thread_size; ++i)
        pthread_join(pool->threads[i], NULL);
    assert(!pthread_mutex_lock(&pool->lock));

    safe_free((void **)&pool->threads);
    safe_free((void **)&pool->functions);
    safe_free((void **)&pool->args);
    pool->thread_size = 0u;
    pool->size        = 0u;
    pool->len         = 0u;

    assert(!pthread_mutex_unlock(&pool->lock));
}

int
worker_pool_push(struct worker_pool *pool, worker_function func, void *arg)
{
    int ret = 1;
    assert(!pthread_mutex_lock(&pool->lock));
    if (pool->exit) {
        LOG_ERROR("WORKER", "Can't push new jobs to worker if interupt was requested");
        goto error;
    }

    if (pool->len == pool->size) {
        if (pool->size < 1)
            pool->size = 2;
        size_t nsize   = 2 * pool->size;
        void *new_func = safe_realloc((void *)pool->functions, nsize * sizeof(worker_function));
        void *new_args = safe_realloc((void *)pool->args, nsize * sizeof(void *));

        pool->size *= 2;
        pool->functions = new_func;
        pool->args      = new_args;
    }

    pool->functions[pool->len] = func;
    pool->args[pool->len]      = arg;
    ++(pool->len);
    ret = 0;
error:
    assert(!pthread_mutex_unlock(&pool->lock));
    return ret;
}

void
worker_pool_waitall(struct worker_pool *pool)
{
    /* No lock, nothing, just test and yield */
    FOR_EVER_IF (pool->len)
        sched_yield();
}

size_t
worker_pool_get_working_count(struct worker_pool *pool)
{
    return pool->thread_working;
}
